"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import pytest
import requests
from datetime import timedelta, datetime
from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.by import By
from pytz import UTC
from io import BytesIO
from sqlalchemy import func
from decimal import Decimal

from biweeklybudget.utils import dtnow
from biweeklybudget.models.account import Account
from biweeklybudget.models.txn_reconcile import TxnReconcile
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.ofxapi import apiclient
from biweeklybudget.ofxapi.exceptions import DuplicateFileException
from ofxparse import OfxParser

fixturedir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', '..', '..', 'fixtures')
)


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestOFX(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/ofx')

    def test_1_heading(self, selenium):
        heading = selenium.find_element(By.CLASS_NAME, 'navbar-brand')
        assert heading.text == 'OFX Transactions - BiweeklyBudget'

    def test_2_nav_menu(self, selenium):
        ul = selenium.find_element(By.ID, 'side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_3_notifications(self, selenium):
        div = selenium.find_element(By.ID, 'notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestOFXDefault(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.dt = dtnow()
        self.get(selenium, base_url + '/ofx')

    def test_table(self, selenium):
        table = selenium.find_element(By.ID, 'table-ofx-txn')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        assert texts[0] == [
            (self.dt - timedelta(hours=13)).strftime('%Y-%m-%d'),
            '-$16.25',
            'CreditOne (3)',
            'debit',
            'INTEREST CHARGED TO STANDARD PUR',
            '38328',
            'CreditOneT3Desc',
            'T3',
            '6',
            (self.dt - timedelta(hours=13)).strftime('%Y-%m-%d'),
            ''
        ]
        assert texts[1] == [
            (self.dt - timedelta(hours=22)).strftime('%Y-%m-%d'),
            '-$123.81',
            'CreditOne (3)',
            'Purchase',
            '123.81 Credit Purchase T1',
            '38328',
            'CreditOneT1Desc',
            'T1',
            '6',
            (self.dt - timedelta(hours=13)).strftime('%Y-%m-%d'),
            ''
        ]
        assert elems[0][2].get_attribute(
            'innerHTML') == '<a href="/accounts/3">CreditOne (3)</a>'
        assert elems[0][7].get_attribute(
            'innerHTML') == '<a href="javascript:ofxTransModal(' \
                            '3, \'T3\')">T3</a>'
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditOne (3)', 'T3'],
            ['CreditOne (3)', 'T1'],
            ['CreditTwo (4)', '001'],
            ['CreditOne (3)', 'T2'],
            ['CreditTwo (4)', '002'],
            ['BankOne (1)', 'BankOne.0.1'],
            ['BankOne (1)', 'BankOne.1.1'],
            ['BankOne (1)', 'BankOne.1.2'],
            ['BankOne (1)', 'BankOne.1.3'],
            ['BankOne (1)', 'BankOne.1.4']
        ]
        reconciles = [t[10] for t in texts]
        assert reconciles == [
            '',
            '',
            '',
            '',
            '',
            'Yes (1)',
            '',
            '',
            '',
            ''
        ]
        assert elems[5][10].get_attribute('innerHTML') == '<a ' \
            'href="javascript:txnReconcileModal(1)">Yes (1)</a>'

    def test_paginate(self, selenium):
        self.get(selenium, self.baseurl + '/ofx')
        table = self.retry_stale(lambda: selenium.find_element(By.ID, 'table-ofx-txn'))
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditOne (3)', 'T3'],
            ['CreditOne (3)', 'T1'],
            ['CreditTwo (4)', '001'],
            ['CreditOne (3)', 'T2'],
            ['CreditTwo (4)', '002'],
            ['BankOne (1)', 'BankOne.0.1'],
            ['BankOne (1)', 'BankOne.1.1'],
            ['BankOne (1)', 'BankOne.1.2'],
            ['BankOne (1)', 'BankOne.1.3'],
            ['BankOne (1)', 'BankOne.1.4']
        ]
        page2_link = selenium.find_element(By.XPATH,
                                           '//li[@class="paginate_button "]/a[text()="2"]'
                                           )
        page2_link.click()
        table = self.retry_stale(lambda: selenium.find_element(By.ID, 'table-ofx-txn'))
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['BankOne (1)', 'BankOne.1.5'],
            ['BankOne (1)', 'BankOne.1.6'],
            ['BankOne (1)', 'BankOne.1.7'],
            ['BankOne (1)', 'BankOne.1.8'],
            ['BankOne (1)', 'BankOne.1.9'],
            ['BankOne (1)', 'BankOne.1.10'],
            ['BankOne (1)', 'BankOne.1.11'],
            ['BankOne (1)', 'BankOne.1.12'],
            ['BankOne (1)', 'BankOne.1.13'],
            ['BankOne (1)', 'BankOne.1.14']
        ]

    def test_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/ofx')
        acct_filter = Select(selenium.find_element(By.ID, 'account_filter'))
        # find the options
        opts = []
        for o in acct_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]

    def test_filter(self, selenium):
        p1trans = [
            ['CreditOne (3)', 'T3'],
            ['CreditOne (3)', 'T1'],
            ['CreditTwo (4)', '001'],
            ['CreditOne (3)', 'T2'],
            ['CreditTwo (4)', '002'],
            ['BankOne (1)', 'BankOne.0.1'],
            ['BankOne (1)', 'BankOne.1.1'],
            ['BankOne (1)', 'BankOne.1.2'],
            ['BankOne (1)', 'BankOne.1.3'],
            ['BankOne (1)', 'BankOne.1.4']
        ]
        self.get(selenium, self.baseurl + '/ofx')
        table = self.retry_stale(lambda: selenium.find_element(By.ID, 'table-ofx-txn'))
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        # check sanity
        assert trans == p1trans
        acct_filter = Select(selenium.find_element(By.ID, 'account_filter'))
        # select CreditTwo (4)
        acct_filter.select_by_value('4')
        table = self.retry_stale(lambda: selenium.find_element(By.ID, 'table-ofx-txn'))
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditTwo (4)', '001'],
            ['CreditTwo (4)', '002']
        ]
        # select back to all
        acct_filter.select_by_value('None')
        table = self.retry_stale(lambda: selenium.find_element(By.ID, 'table-ofx-txn'))
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == p1trans

    def test_search(self, selenium):
        p1trans = [
            ['CreditOne (3)', 'T3'],
            ['CreditTwo (4)', '001'],
            ['BankTwoStale (2)', '1'],
            ['CreditOne (3)', 'T2-2'],
            ['DisabledBank (6)', '001']
        ]
        self.get(selenium, self.baseurl + '/ofx')
        search = self.retry_stale(
            lambda: selenium.find_element(By.XPATH, '//input[@type="search"]')
        )
        search.send_keys('inte')
        table = self.retry_stale(lambda: selenium.find_element(By.ID, 'table-ofx-txn'))
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        # check sanity
        assert trans == p1trans
        acct_filter = Select(selenium.find_element(By.ID, 'account_filter'))
        # select CreditTwo (4)
        acct_filter.select_by_value('4')
        table = self.retry_stale(lambda: selenium.find_element(By.ID, 'table-ofx-txn'))
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditTwo (4)', '001']
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestOFXTransModal(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/ofx')

    def test_modal_on_click(self, selenium):
        link = selenium.find_element(By.XPATH, '//a[text()="T1"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'OFXTransaction Account=3 FITID=T1'
        texts = self.tbody2textlist(body)
        elems = self.tbody2elemlist(body)
        pdate = (dtnow() - timedelta(hours=22))
        fdate = (dtnow() - timedelta(hours=13))
        assert texts == [
            ['Account', 'CreditOne (3)'],
            ['FITID', 'T1'],
            ['Date Posted', pdate.strftime('%Y-%m-%d')],
            ['Amount', '-$123.81'],
            ['Name', '123.81 Credit Purchase T1'],
            ['Memo', '38328'],
            ['Type', 'Purchase'],
            ['Description', 'CreditOneT1Desc'],
            ['Notes', ''],
            ['Checknum', '123'],
            ['MCC', 'T1mcc'],
            ['SIC', 'T1sic'],
            ['OFX Statement'],
            ['ID', '6'],
            ['Date', fdate.strftime('%Y-%m-%d')],
            ['Filename', '/stmt/CreditOne/0'],
            ['File mtime', fdate.strftime('%Y-%m-%d')],
            ['Ledger Balance', '-$952.06']
        ]
        assert elems[0][1].get_attribute(
            'innerHTML') == '<a href="/accounts/3">CreditOne (3)</a>'
        assert selenium.find_element(By.ID,
                                     'modalSaveButton').is_displayed() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestOFXTransURL(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/ofx/3/T1')

    def test_modal_auto_displayed(self, selenium):
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'OFXTransaction Account=3 FITID=T1'
        texts = self.tbody2textlist(body)
        elems = self.tbody2elemlist(body)
        pdate = (dtnow() - timedelta(hours=22))
        fdate = (dtnow() - timedelta(hours=13))
        assert texts == [
            ['Account', 'CreditOne (3)'],
            ['FITID', 'T1'],
            ['Date Posted', pdate.strftime('%Y-%m-%d')],
            ['Amount', '-$123.81'],
            ['Name', '123.81 Credit Purchase T1'],
            ['Memo', '38328'],
            ['Type', 'Purchase'],
            ['Description', 'CreditOneT1Desc'],
            ['Notes', ''],
            ['Checknum', '123'],
            ['MCC', 'T1mcc'],
            ['SIC', 'T1sic'],
            ['OFX Statement'],
            ['ID', '6'],
            ['Date', fdate.strftime('%Y-%m-%d')],
            ['Filename', '/stmt/CreditOne/0'],
            ['File mtime', fdate.strftime('%Y-%m-%d')],
            ['Ledger Balance', '-$952.06']
        ]
        assert elems[0][1].get_attribute(
            'innerHTML') == '<a href="/accounts/3">CreditOne (3)</a>'
        assert selenium.find_element(By.ID,
                                     'modalSaveButton').is_displayed() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
@pytest.mark.incremental
class TestTransReconciledModal(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(TxnReconcile).get(1)
        assert t.ofx_account_id == 1
        assert t.ofx_fitid == 'BankOne.0.1'
        assert t.txn_id == 1
        assert t.rule_id is None
        assert t.note == 'reconcile notes'
        assert t.reconciled_at == datetime(2017, 4, 10, 8, 9, 11, tzinfo=UTC)

    def test_1_modal(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element(By.XPATH,
                                     '//a[@href="javascript:txnReconcileModal(1)"]')
        modal, title, body = self.try_click_and_get_modal(selenium, link)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Transaction Reconcile 1'
        dl = body.find_element(By.TAG_NAME, 'dl')
        assert dl.get_attribute('innerHTML') == '\n' \
            '<dt>Date Reconciled</dt><dd>2017-04-10 08:09:11 UTC</dd>\n' \
            '<dt>Note</dt><dd>reconcile notes</dd>\n' \
            '<dt>Rule</dt><dd>null</dd>\n'
        trans_tbl = body.find_element(By.ID, 'txnReconcileModal-trans')
        trans_texts = self.tbody2textlist(trans_tbl)
        assert trans_texts == [
            ['Transaction'],
            [
                'Date',
                (dtnow() + timedelta(days=4)).strftime('%Y-%m-%d')
            ],
            ['Amount', '$111.13'],
            ['Budgeted Amount', '$111.11'],
            ['Description', 'T1foo'],
            ['Account', 'BankOne (1)'],
            ['Budget', 'Periodic1 (1)'],
            ['Notes', 'notesT1'],
            ['Scheduled?', 'Yes (1)']
        ]
        trans_elems = self.tbody2elemlist(trans_tbl)
        assert trans_elems[5][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'
        assert trans_elems[6][1].get_attribute('innerHTML') == '<a href=' \
            '"/budgets/1">Periodic1 (1)</a>'
        assert trans_elems[8][1].get_attribute('innerHTML') == '<a href=' \
            '"/scheduled/1">Yes (1)</a>'
        ofx_tbl = body.find_element(By.ID, 'txnReconcileModal-ofx')
        ofx_texts = self.tbody2textlist(ofx_tbl)
        assert ofx_texts == [
            ['OFX Transaction'],
            ['Account', 'BankOne (1)'],
            ['FITID', 'BankOne.0.1'],
            ['Date Posted', (dtnow() - timedelta(days=6)).strftime('%Y-%m-%d')],
            ['Amount', '-$20.00'],
            ['Name', 'Late Fee'],
            ['Memo', ''],
            ['Type', 'Debit'],
            ['Description', ''],
            ['Notes', ''],
            ['Checknum', ''],
            ['MCC', ''],
            ['SIC', ''],
            ['OFX Statement'],
            ['ID', '1'],
            ['Date', (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')],
            ['Filename', '/stmt/BankOne/0'],
            [
                'File mtime',
                (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')
            ],
            ['Ledger Balance', '$12,345.67']
        ]
        ofx_elems = self.tbody2elemlist(ofx_tbl)
        assert ofx_elems[1][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
@pytest.mark.skip('Deprecated, will be removed')
class TestOfxApi(AcceptanceHelper):

    def test_0_set_account_regexes(self, testdb):
        acct = testdb.query(Account).get(3)
        acct.re_interest_paid = None
        acct.re_payment = 'INTERNET PAYMENT - THANK.*'
        testdb.commit()
        assert acct.re_interest_charge == '^INTEREST CHARGED TO'
        assert acct.re_interest_paid is None
        assert acct.re_payment == 'INTERNET PAYMENT - THANK.*'
        assert acct.re_late_fee == '^Late Fee'
        assert acct.re_other_fee == '^re-other-fee'

    def test_1_verify_db(self, testdb):
        assert testdb.query(OFXStatement).with_entities(
            func.max(OFXStatement.id)
        ).scalar() == 9
        assert testdb.query(OFXTransaction).with_entities(
            func.max(OFXTransaction.statement_id)
        ).scalar() == 9
        assert len(testdb.query(OFXTransaction).all()) == 33
        acct = testdb.query(Account).get(3)
        assert acct.re_interest_charge == '^INTEREST CHARGED TO'
        assert acct.re_interest_paid is None
        assert acct.re_payment == 'INTERNET PAYMENT - THANK.*'
        assert acct.re_late_fee == '^Late Fee'
        assert acct.re_other_fee == '^re-other-fee'

    def test_2_get_accounts(self, base_url):
        r = requests.get(base_url + '/api/ofx/accounts')
        assert r.status_code == 200
        assert r.json() == {
            'BankOne': {
                'cat_memo': True,
                'config': {'foo': 'bar'},
                'id': 1,
                'vault_path': 'secret/foo/bar/BankOne'
            },
            'BankTwoStale': {
                'cat_memo': False,
                'config': {'foo': 'baz'},
                'id': 2,
                'vault_path': 'secret/foo/bar/BankTwo'
            },
            'CreditTwo': {
                'cat_memo': False,
                'config': {},
                'id': 4,
                'vault_path': '/foo/bar'
            },
            'DisabledBank': {
                'cat_memo': True,
                'config': {'bar': 'baz'},
                'id': 6,
                'vault_path': ''
            },
            'InvestmentOne': {
                'cat_memo': False,
                'config': {},
                'id': 5,
                'vault_path': ''
            }
        }

    def test_3_post_ofx(self, base_url):
        ofxpath = os.path.join(fixturedir, 'CreditOne_2017-07-28_05-30-00.ofx')
        with open(ofxpath, 'rb') as fh:
            ofx_str = fh.read()
        ofx = OfxParser.parse(BytesIO(ofx_str))
        client = apiclient(base_url)
        stmt_id, count_new, count_upd = client.update_statement_ofx(
            3, ofx, filename='/statements/CreditOne/'
                             'CreditOne_2017-07-28_05-30-00.ofx'
        )
        assert stmt_id == 10
        assert count_new == 1
        assert count_upd == 0

    def test_4_verify_db(self, testdb):
        assert testdb.query(OFXStatement).with_entities(
            func.max(OFXStatement.id)
        ).scalar() == 10
        assert testdb.query(OFXTransaction).with_entities(
            func.max(OFXTransaction.statement_id)
        ).scalar() == 10
        assert len(testdb.query(OFXTransaction).all()) == 34
        stmt = testdb.query(OFXStatement).get(10)
        assert stmt.id == 10
        assert stmt.account_id == 3
        assert stmt.filename == '/statements/CreditOne/' \
                                'CreditOne_2017-07-28_05-30-00.ofx'
        assert stmt.file_mtime == dtnow()
        assert stmt.currency == 'USD'
        assert stmt.bankid == '4321'
        assert stmt.routing_number == ''
        assert stmt.acct_type == ''
        assert stmt.brokerid is None
        assert stmt.acctid == 'CreditOneAcctId'
        assert stmt.type == 'CreditCard'
        assert stmt.as_of == datetime(2017, 7, 28, 5, 30, 0, tzinfo=UTC)
        assert stmt.ledger_bal == Decimal('-1234.5600')
        assert stmt.ledger_bal_as_of == datetime(
            2017, 7, 28, 5, 29, 32, tzinfo=UTC
        )
        assert stmt.avail_bal is None
        assert stmt.avail_bal_as_of is None
        trans = testdb.query(OFXTransaction).get((3, 'FITID20170727144.0G53TY'))
        assert trans.account_id == 3
        assert trans.statement_id == 10
        assert trans.fitid == 'FITID20170727144.0G53TY'
        assert trans.trans_type == 'credit'
        assert trans.date_posted == datetime(2017, 7, 27, 16, 0, 0, tzinfo=UTC)
        assert trans.amount == Decimal('123.0000')
        assert trans.name == 'INTERNET PAYMENT - THANK YOU'
        assert trans.memo == ''
        assert trans.sic is None
        assert trans.mcc == ''
        assert trans.checknum is None
        assert trans.description is None
        assert trans.notes is None
        assert trans.reconcile_id is None
        assert trans.is_interest_charge is False
        assert trans.is_interest_payment is False
        assert trans.is_late_fee is False
        assert trans.is_other_fee is False
        assert trans.is_payment is True

    def test_5_post_same_ofx(self, base_url):
        ofxpath = os.path.join(fixturedir, 'CreditOne_2017-07-28_05-30-00.ofx')
        with open(ofxpath, 'rb') as fh:
            ofx_str = fh.read()
        ofx = OfxParser.parse(BytesIO(ofx_str))
        client = apiclient(base_url)
        with pytest.raises(DuplicateFileException) as ex:
            client.update_statement_ofx(
                3, ofx, filename='/statements/CreditOne/'
                                 'CreditOne_2017-07-28_05-30-00.ofx'
            )
        assert ex.value.acct_id == 3
        assert ex.value.stmt_id == 10
        assert ex.value.filename == '/statements/CreditOne/' \
                                    'CreditOne_2017-07-28_05-30-00.ofx'



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from sqlalchemy import (
    Column, Integer, String, ForeignKey, Numeric, UniqueConstraint
)
from sqlalchemy_utc import UtcDateTime
from sqlalchemy.orm import relationship
from biweeklybudget.models.base import Base, ModelAsDict


class OFXStatement(Base, ModelAsDict):

    __tablename__ = 'ofx_statements'
    __table_args__ = (
        UniqueConstraint('account_id', 'filename'),
        {'mysql_engine': 'InnoDB'}
    )

    #: Unique ID
    id = Column(Integer, primary_key=True)

    #: Foreign key - Account.id - ID of the account this statement is for
    account_id = Column(Integer, ForeignKey('accounts.id'), nullable=False)

    #: Relationship to the :py:class:`~.Account` this statement is for
    account = relationship(
        "Account", uselist=False
    )

    #: Filename parsed from
    filename = Column(String(254))

    #: File mtime
    file_mtime = Column(UtcDateTime)

    #: Currency definition ("USD")
    currency = Column(String(10))

    #: FID of the Institution
    bankid = Column(String(20))

    #: Routing Number
    routing_number = Column(String(20))

    #: Textual account type, from the bank (i.e. "Checking")
    acct_type = Column(String(32))

    #: BrokerID, for investment accounts
    brokerid = Column(String(30))

    #: Institution's account ID
    acctid = Column(String(30))

    #: Account Type, string corresponding to ofxparser.ofxparser.AccountType
    type = Column(String(20))

    #: Last OFX statement datetime
    as_of = Column(UtcDateTime)

    #: Ledger balance, or investment account value
    ledger_bal = Column(Numeric(precision=10, scale=4))

    #: as-of date for the ledger balance
    ledger_bal_as_of = Column(UtcDateTime)

    #: Available balance
    avail_bal = Column(Numeric(precision=10, scale=4))

    #: as-of date for the available balance
    avail_bal_as_of = Column(UtcDateTime)

    def __repr__(self):
        return "<OFXStatement(id=%s, account_id=%s, as_of='%s')>" % (
            self.id, self.account_id, self.as_of
        )



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify, request
from datatables import DataTable
from copy import copy
from datetime import datetime
from decimal import Decimal

from biweeklybudget.db import db_session
from biweeklybudget.flaskapp.app import app
from biweeklybudget.models.account import Account
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.models.scheduled_transaction import ScheduledTransaction
from biweeklybudget.flaskapp.views.searchableajaxview import SearchableAjaxView
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView

logger = logging.getLogger(__name__)


class ScheduledView(MethodView):

    def get(self):
        """
        Render the GET /scheduled view using the ``scheduled.html`` template.
        """
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        for b in db_session.query(Budget).all():
            if b.is_income:
                budgets['%s (income)' % b.name] = b.id
            else:
                budgets[b.name] = b.id
        return render_template(
            'scheduled.html',
            accts=accts,
            budgets=budgets
        )


class ScheduledTransView(MethodView):

    def get(self, sched_trans_id):
        """
        Render the GET /scheduled/<int:sched_trans_id> view using the
        ``scheduled.html`` template.
        """
        accts = {a.name: a.id for a in db_session.query(Account).all()}
        budgets = {}
        for b in db_session.query(Budget).all():
            if b.is_income:
                budgets['%s (income)' % b.name] = b.id
            else:
                budgets[b.name] = b.id
        return render_template(
            'scheduled.html',
            accts=accts,
            budgets=budgets,
            sched_trans_id=sched_trans_id
        )


class ScheduledAjax(SearchableAjaxView):
    """
    Handle GET /ajax/scheduled endpoint.
    """

    def _filterhack(self, qs, s, args):
        """
        DataTables 1.10.12 has built-in support for filtering based on a value
        in a specific column; when this is done, the filter value is set in
        ``columns[N][search][value]`` where N is the column number. However,
        the python datatables package used here only supports the global
        ``search[value]`` input, not the per-column one.

        However, the DataTable search is implemented by passing a callable
        to ``table.searchable()`` which takes two arguments, the current Query
        that's being built, and the user's ``search[value]`` input; this must
        then return a Query object with the search applied.

        In python datatables 0.4.9, this code path is triggered on
        ``if callable(self.search_func) and search.get("value", None):``

        As such, we can "trick" the table to use per-column searching (currently
        only if global searching is not being used) by examining the per-column
        search values in the request, and setting the search function to one
        (this method) that uses those values instead of the global
        ``search[value]``.

        :param qs: Query currently being built
        :type qs: ``sqlalchemy.orm.query.Query``
        :param s: user search value
        :type s: str
        :param args: args
        :type args: dict
        :return: Query with searching applied
        :rtype: ``sqlalchemy.orm.query.Query``
        """
        # Ok, build our filter...
        type_filter = args['columns'][1]['search']['value']
        if type_filter != '' and type_filter != 'None':
            qs = qs.filter(
                ScheduledTransaction.schedule_type.__eq__(type_filter)
            )
        # search
        if s != '' and s != 'FILTERHACK':
            if len(s) < 3:
                return qs
            s = '%' + s + '%'
            qs = qs.filter(ScheduledTransaction.description.like(s))
        return qs

    def get(self):
        """
        Render and return JSON response for GET /ajax/ofx
        """
        args = request.args.to_dict()
        args_dict = self._args_dict(args)
        if self._have_column_search(args_dict) and args['search[value]'] == '':
            args['search[value]'] = 'FILTERHACK'
        table = DataTable(
            args, ScheduledTransaction, db_session.query(ScheduledTransaction),
            [
                'is_active',
                (
                    'amount',
                    lambda a: float(a.amount)
                ),
                (
                    'sales_tax',
                    lambda a: float(a.sales_tax)
                ),
                'description',
                (
                    'account',
                    'account.name',
                    lambda i: "{} ({})".format(i.name, i.id)
                ),
                (
                    'budget',
                    'budget.name',
                    lambda i: "{} {}({})".format(
                        i.name,
                        '(income) ' if i.is_income else '',
                        i.id
                    )
                ),
                'recurrence_str',
                'schedule_type'
            ]
        )
        table.add_data(
            acct_id=lambda o: o.account_id,
            budget_id=lambda o: o.budget_id,
            id=lambda o: o.id
        )
        if args['search[value]'] != '':
            table.searchable(lambda qs, s: self._filterhack(qs, s, args_dict))
        return jsonify(table.json())


class OneScheduledAjax(MethodView):
    """
    Handle GET /ajax/scheduled/<int:sched_trans_id> endpoint.
    """

    def get(self, sched_trans_id):
        t = db_session.query(ScheduledTransaction).get(sched_trans_id)
        d = copy(t.as_dict)
        d['account_name'] = t.account.name
        d['budget_name'] = t.budget.name
        return jsonify(d)


class SchedTransFormHandler(FormHandlerView):
    """
    Handle POST /forms/scheduled
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data.get('description', '').strip() == '':
            errors['description'].append('Description cannot be empty')
            have_errors = True
        if float(data['amount']) == 0:
            errors['amount'].append('Amount cannot be zero')
            have_errors = True
        if data['account'] == 'None':
            errors['account'].append('Transactions must have an account')
            have_errors = True
        if data['budget'] == 'None':
            errors['budget'].append('Transactions must have a budget')
            have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        if 'id' in data and data['id'].strip() != '':
            # updating an existing budget
            trans = db_session.query(ScheduledTransaction).get(int(data['id']))
            if trans is None:
                raise RuntimeError("Error: no ScheduledTransaction with ID "
                                   "%s" % data['id'])
            action = 'updating ScheduledTransaction ' + data['id']
        else:
            trans = ScheduledTransaction()
            action = 'creating new ScheduledTransaction'
        trans.description = data['description'].strip()
        # Clear all schedule type fields first, then set the appropriate one
        trans.date = None
        trans.day_of_month = None
        trans.num_per_period = None
        trans.day_of_week = None
        trans.annual_month = None
        trans.annual_day = None
        if data['type'] == 'monthly':
            trans.day_of_month = int(data['day_of_month'])
        elif data['type'] == 'per_period':
            trans.num_per_period = int(data['num_per_period'])
        elif data['type'] == 'weekly':
            trans.day_of_week = int(data['day_of_week'])
        elif data['type'] == 'annual':
            trans.annual_month = int(data['annual_month'])
            trans.annual_day = int(data['annual_day'])
        else:
            # date
            trans.date = datetime.strptime(data['date'], '%Y-%m-%d').date()
        trans.amount = Decimal(data['amount'])
        if data.get('sales_tax', '').strip() == '':
            trans.sales_tax = Decimal('0.0')
        else:
            trans.sales_tax = Decimal(data['sales_tax'])
        trans.account_id = int(data['account'])
        trans.budget_id = int(data['budget'])
        trans.notes = data['notes'].strip()
        trans.is_active = data['is_active']
        logger.info('%s: %s', action, trans.as_dict)
        db_session.add(trans)
        db_session.commit()
        return 'Successfully saved ScheduledTransaction' \
               ' %d in database.' % trans.id


app.add_url_rule(
    '/scheduled',
    view_func=ScheduledView.as_view('scheduled_view')
)
app.add_url_rule(
    '/scheduled/<int:sched_trans_id>',
    view_func=ScheduledTransView.as_view('sched_trans')
)
app.add_url_rule(
    '/ajax/scheduled',
    view_func=ScheduledAjax.as_view('scheduled_ajax')
)
app.add_url_rule(
    '/ajax/scheduled/<int:sched_trans_id>',
    view_func=OneScheduledAjax.as_view('one_scheduled_ajax')
)
app.add_url_rule(
    '/forms/scheduled',
    view_func=SchedTransFormHandler.as_view('scheduled_form')
)



"""
account interest and payoff attributes

Revision ID: 9dc8545963be
Revises: ceb73ddf66eb
Create Date: 2017-08-12 21:27:44.820078

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9dc8545963be'
down_revision = 'ceb73ddf66eb'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        'accounts',
        sa.Column(
            'apr',
            sa.Numeric(precision=5, scale=4),
            nullable=True
        )
    )
    op.add_column(
        'accounts',
        sa.Column(
            'prime_rate_margin',
            sa.Numeric(precision=5, scale=4),
            nullable=True
        )
    )
    op.add_column(
        'accounts',
        sa.Column(
            'interest_class_name',
            sa.String(length=70),
            nullable=True
        )
    )
    op.add_column(
        'accounts',
        sa.Column(
            'min_payment_class_name',
            sa.String(length=70),
            nullable=True
        )
    )


def downgrade():
    op.drop_column('accounts', 'min_payment_class_name')
    op.drop_column('accounts', 'interest_class_name')
    op.drop_column('accounts', 'prime_rate_margin')
    op.drop_column('accounts', 'apr')



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from decimal import Decimal
import json
from datetime import timedelta

import lxml.html
from lxml.etree import tostring
import requests
import logging

from biweeklybudget.models.dbsetting import DBSetting
from biweeklybudget.utils import dtnow, decode_json_datetime
from biweeklybudget.flaskapp.jsonencoder import MagicJSONEncoder

logger = logging.getLogger(__name__)


class PrimeRateCalculator(object):

    def __init__(self, db_session):
        """
        :param db_session: Database session
        :type db_session: sqlalchemy.orm.session.Session
        """
        self._db_sess = db_session

    def _rate_from_marketwatch(self):
        url = 'https://www.wsj.com/market-data/bonds/moneyrates'
        logger.debug('Requesting %s for prime rate', url)
        r = requests.get(
            url,
            headers={'User-Agent': 'Mozilla/5.0 (X11; Linux x86_64) '
                                   'AppleWebKit/537.36 (KHTML, like Gecko) '
                                   'Chrome/79.0.3945.130 Safari/537.36'}
        )
        doc = lxml.html.fromstring(r.text)
        prtitle = doc.xpath('//caption[starts-with(text(), "Prime Rates")]')
        if prtitle is None or len(prtitle) != 1:
            logger.error(
                'Could not find "Prime rates" TD; page: %s', r.text
            )
            raise RuntimeError('ERROR: format change in %s' % url)
        prtbody = prtitle[0].getparent().findall('tbody')[0]
        prthead = prtitle[0].getparent().findall('thead')[0]
        headrows = prthead.findall('tr')
        if not headrows[1].findall('th')[1].text == 'Latest':
            logger.error(
                'tr2,td1 contains unexpected text; prtbody: %s',
                tostring(prtbody)
            )
            raise RuntimeError('ERROR: format change in %s' % url)
        rows = prtbody.findall('tr')
        if not rows[0].findall('td')[0].text == 'U.S.':
            logger.error(
                'tr3,td0 contains unexpected text; prtbody: %s',
                tostring(prtbody)
            )
            raise RuntimeError('ERROR: format change in %s' % url)
        pr_amt = rows[0].findall('td')[1].text
        logger.debug(
            'Found prime rate from %s: %s', url, pr_amt
        )
        return pr_amt

    def _get_prime_rate(self):
        """
        Get the US Prime Rate from MarketWatch; update the DB and return the
        value.

        :return: current US Prime Rate
        :rtype: decimal.Decimal
        """
        rate = self._rate_from_marketwatch()
        rate = Decimal(rate) * Decimal('0.01')
        s = self._db_sess.query(DBSetting).get('prime_rate')
        if s is None:
            s = DBSetting(name='prime_rate')
        logger.info('Got Prime Rate from MarketWatch: %s', rate)
        s.value = json.dumps({
            'value': '%s' % rate,
            'date': dtnow()
        }, cls=MagicJSONEncoder)
        self._db_sess.add(s)
        self._db_sess.flush()
        self._db_sess.commit()
        return rate

    @property
    def prime_rate(self):
        """
        Return the current US Prime Rate

        :return: current US Prime Rate
        :rtype: decimal.Decimal
        """
        pr = self._db_sess.query(DBSetting).get('prime_rate')
        if pr is None:
            return self._get_prime_rate()
        j = json.loads(pr.value)
        d = decode_json_datetime(j['date'])
        if d >= (dtnow() - timedelta(hours=48)):
            return Decimal(j['value'])
        return self._get_prime_rate()

    def calculate_apr(self, margin):
        """
        Calculate an APR based on the prime rate.

        :param margin: margin added to Prime Rate to get APR
        :type margin: decimal.Decimal
        :return: effective APR
        :rtype: decimal.Decimal
        """
        return self.prime_rate + margin



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from decimal import Decimal
from sqlalchemy import (
    Column, Integer, String, ForeignKey, Numeric, Boolean, inspect, func
)
from sqlalchemy.orm import relationship
from sqlalchemy.ext.hybrid import hybrid_property
from biweeklybudget.models.base import Base, ModelAsDict

logger = logging.getLogger(__name__)


class Project(Base, ModelAsDict):

    __tablename__ = 'projects'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Name of project
    name = Column(String(40))

    #: Notes / Description
    notes = Column(String(254))

    #: whether active or historical
    is_active = Column(Boolean, default=True)

    def __repr__(self):
        return "<Project(id=%s, name=%s)>" % (
            self.id, self.name
        )

    @property
    def total_cost(self):
        """
        Return the total cost of all line items (:py:class:`~.BoMItem`) for
        this project.

        :return: total cost of this project
        :rtype: float
        """
        sess = inspect(self).session
        cost = sess.query(BoMItem).filter(
            BoMItem.project_id.__eq__(self.id)
        ).with_entities(
            func.sum(BoMItem.line_cost)
        ).scalar()
        if cost is None:
            return Decimal('0.0')
        return cost

    @property
    def remaining_cost(self):
        """
        Return the remaining cost of all line items (:py:class:`~.BoMItem`) for
        this project which are still active

        :return: remianing cost of this project
        :rtype: float
        """
        sess = inspect(self).session
        cost = sess.query(BoMItem).filter(
            BoMItem.project_id.__eq__(self.id),
            BoMItem.is_active.__eq__(True)
        ).with_entities(
            func.sum(BoMItem.line_cost)
        ).scalar()
        if cost is None:
            return Decimal('0.0')
        return cost


class BoMItem(Base, ModelAsDict):
    __tablename__ = 'bom_items'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: Project ID
    project_id = Column(Integer, ForeignKey('projects.id'), nullable=False)

    #: Relationship to the :py:class:`~.Project` this item is for
    project = relationship('Project', uselist=False)

    #: Name of item
    name = Column(String(254))

    #: Notes / Description
    notes = Column(String(254))

    #: Quantity Required
    quantity = Column(Integer, default=1)

    #: Unit Cost / Cost Each
    unit_cost = Column(Numeric(precision=10, scale=4), default=0.0)

    #: URL
    url = Column(String(254))

    #: whether active or historical
    is_active = Column(Boolean, default=True)

    def __repr__(self):
        return "<BoMItem(id=%s, name=%s, project_id=%s)>" % (
            self.id, self.name, self.project_id
        )

    @hybrid_property
    def line_cost(self):
        """
        The total cost for this BoM Item, unit_cost times quantity

        :return: total line cost
        :rtype: decimal.Decimal
        """
        return (self.quantity * Decimal(1.0)) * self.unit_cost



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import logging

import plaid

from biweeklybudget import settings
from datetime import datetime
import pytz
from contextlib import contextmanager
from babel.numbers import format_currency
from plaid import Configuration, ApiClient, Environment
from plaid.api.plaid_api import PlaidApi

logger = logging.getLogger(__name__)


def plaid_client() -> PlaidApi:
    """
    Return an initialized Plaid API client instance.
    """
    logger.debug('Getting Plaid client instance')
    assert settings.PLAID_CLIENT_ID is not None
    assert settings.PLAID_SECRET is not None
    assert settings.PLAID_ENV is not None
    configuration = Configuration(
        host=getattr(Environment, settings.PLAID_ENV),
        api_key={
            'clientId': settings.PLAID_CLIENT_ID,
            'secret': settings.PLAID_SECRET,
        }
    )
    api_client = ApiClient(configuration)
    client = PlaidApi(api_client)
    return client


def fix_werkzeug_logger():
    """
    Remove the werkzeug logger StreamHandler (call from ``app.py``).

    With Werkzeug at least as of 0.12.1, werkzeug._internal._log sets up its own
    StreamHandler if logging isn't already configured. Because we're using
    the ``flask`` command line wrapper, that will ALWAYS be imported (and
    executed) before we can set up our own logger. As a result, to fix the
    duplicate log messages, we have to go back and remove that StreamHandler.
    """
    wlog = logging.getLogger('werkzeug')
    logger.info('Removing handlers from "werkzeug" logger')
    for h in wlog.handlers:
        wlog.removeHandler(h)


def fmt_currency(amt):
    """
    Using :py:attr:`~biweeklybudget.settings.LOCALE_NAME` and
    :py:attr:`~biweeklybudget.settings.CURRENCY_CODE`, return ``amt`` formatted
    as currency.

    :param amt: The amount to format; any numeric type.
    :return: ``amt`` formatted for the appropriate locale and currency
    :rtype: str
    """
    return format_currency(
        amt, settings.CURRENCY_CODE, locale=settings.LOCALE_NAME
    )


def dtnow():
    """
    Return the current datetime as a timezone-aware DateTime object in UTC.

    :return: current datetime
    :rtype: datetime.datetime
    """
    # This is for acceptance tests... :(
    if settings.BIWEEKLYBUDGET_TEST_TIMESTAMP is not None:
        return datetime.fromtimestamp(
            int(settings.BIWEEKLYBUDGET_TEST_TIMESTAMP), pytz.utc
        )
    return datetime.utcnow().replace(tzinfo=pytz.utc)


def decode_json_datetime(d):
    """
    Return a datetime.datetime for a datetime that was serialized with
    :py:class:`~.MagicJSONEncoder`.

    :param d: dict from deserialized JSON
    :type d: dict
    :return: datetime represented by dict
    :rtype: datetime.datetime
    """
    return datetime(
        d['year'], d['month'], d['date'], d['hour'], d['minute'], d['second'],
        tzinfo=pytz.timezone(d['tzname'])
    )


def date_suffix(n):
    """
    Given an integer day of month (1 <= n <= 31), return that number with the
    appropriate suffix (st|nd|rd|th).

    From: http://stackoverflow.com/a/5891598/211734

    :param n: Integer day of month
    :type n: int
    :return: n with the appropriate suffix
    :rtype: str
    """
    return 'th' if 11 <= n <= 13 else {1: 'st', 2: 'nd', 3: 'rd'}.get(
        n % 10, 'th'
    )


@contextmanager
def in_directory(path):
    pwd = os.getcwd()
    os.chdir(path)
    yield os.path.abspath(path)
    os.chdir(pwd)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016-2024 Jason Antman <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from sqlalchemy import (
    Column, Integer, Numeric, String, Date, ForeignKey, inspect, func, select
)
from sqlalchemy.orm import relationship
from sqlalchemy.sql.expression import null
from sqlalchemy.ext.hybrid import hybrid_property
from biweeklybudget.models.base import Base, ModelAsDict
from biweeklybudget.models.budget_transaction import BudgetTransaction
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.utils import dtnow
from biweeklybudget.settings import RECONCILE_BEGIN_DATE

logger = logging.getLogger(__name__)


class Transaction(Base, ModelAsDict):
    """
    Class that describes Transactions that have actually occurred, against one
    account and one or more budgets.

    Note that in addition to the usual class attributes, the constructor of
    this class also accepts a ``budget_amounts`` keyword argument, which passes
    its value on to :py:meth:`~.set_budget_amounts`.
    """

    __tablename__ = 'transactions'
    __table_args__ = (
        {'mysql_engine': 'InnoDB'}
    )

    #: Class properties to include in :py:attr:`~.ModelAsDict.as_dict` result.
    _dict_properties = ['actual_amount']

    #: Primary Key
    id = Column(Integer, primary_key=True)

    #: date of the transaction
    date = Column(Date, default=dtnow().date())

    #: Budgeted amount of the transaction, if it was budgeted ahead of time
    #: via a :py:class:`~.ScheduledTransaction`. This attribute is only set by
    #: :py:meth:`~.SchedToTransFormHandler.submit` and
    #: :py:meth:`~.SkipSchedTransFormHandler.submit`. And, for some incorrect
    #: reason, by :py:func:`biweeklybudget.models.utils.do_budget_transfer`.
    budgeted_amount = Column(Numeric(precision=10, scale=4))

    #: description
    description = Column(String(254), nullable=False, index=True)

    #: free-form notes
    notes = Column(String(254))

    #: ID of the account this transaction is against
    account_id = Column(Integer, ForeignKey('accounts.id'))

    #: Relationship - :py:class:`~.Account` this transaction is against
    account = relationship(
        "Account", backref="transactions", uselist=False
    )

    #: ID of the ScheduledTransaction this Transaction was created from;
    #: set when a scheduled transaction is converted to a real one
    scheduled_trans_id = Column(
        Integer, ForeignKey('scheduled_transactions.id')
    )

    #: Relationship - the :py:class:`~.ScheduledTransaction`
    #: this Transaction was created from; set when a scheduled transaction
    #: is converted to a real one
    scheduled_trans = relationship(
        "ScheduledTransaction", backref="transactions", uselist=False
    )

    #: ID of the Budget this transaction was planned to be funded by, if it
    #: was planned ahead via a :py:class:`~.ScheduledTransaction`
    planned_budget_id = Column(Integer, ForeignKey('budgets.id'))

    #: Relationship - the :py:class:`~.Budget` this transaction was planned to
    #: be funded by, if it was planned ahead via a
    #: :py:class:`~.ScheduledTransaction`.
    planned_budget = relationship(
        "Budget", backref="planned_transactions", uselist=False,
        foreign_keys=[planned_budget_id]
    )

    #: If the transaction is one half of a transfer, the Transaction ID of the
    #: other half/side of the transfer.
    transfer_id = Column(Integer, ForeignKey('transactions.id'))

    #: Relationship - the :py:class:`~.Transaction` that makes up the other
    #: half/side of a transfer, if this transaction was for a transfer.
    transfer = relationship(
        "Transaction", remote_side=[id], post_update=True, uselist=False
    )

    #: Amount of sales tax paid on this transaction.
    sales_tax = Column(
        Numeric(precision=10, scale=4), nullable=False, default=0.0
    )

    def __init__(self, **kwargs):
        """
        Custom constructor for Transaction class to allow setting/syncing
        BudgetTransactions via constructor.

        :param kwargs: class constructor keyword arguments
        :type kwargs: dict
        """
        cls_ = type(self)
        for k in kwargs:
            if k == 'budget_amounts':
                self.set_budget_amounts(kwargs[k])
            elif not hasattr(cls_, k):
                raise TypeError(
                    "%r is an invalid keyword argument for %s" %
                    (k, cls_.__name__))
            setattr(self, k, kwargs[k])

    def __repr__(self):
        return "<Transaction(id=%s)>" % self.id

    @hybrid_property
    def actual_amount(self):
        """
        Actual amount of the transaction.

        :return: actual total amount of the transaction
        :rtype: decimal.Decimal
        """
        return sum([
            bt.amount for bt in self.budget_transactions
        ])

    @actual_amount.expression
    def actual_amount(cls):
        # see: http://docs.sqlalchemy.org/en/latest/orm/extensions/hybrid.html
        # #correlated-subquery-relationship-hybrid
        return select(
            func.sum(BudgetTransaction.amount)
        ).where(
            BudgetTransaction.trans_id.__eq__(cls.id)
        ).label('actual_amount')

    @staticmethod
    def unreconciled(db):
        """
        Return a query to match all unreconciled Transactions.

        :param db: active database session to use for queries
        :type db: sqlalchemy.orm.session.Session
        :return: query to match all unreconciled Transactions
        :rtype: sqlalchemy.orm.query.Query
        """
        return db.query(Transaction).filter(
            Transaction.reconcile.__eq__(null()),
            Transaction.date.__ge__(RECONCILE_BEGIN_DATE),
            Transaction.account.has(reconcile_trans=True)
        )

    def set_budget_amounts(self, budget_amounts):
        """
        Manage child :py:class:`~.BudgetTransaction` objects corresponding to
        budget allocations of the amount of this transaction. Given a dictionary
        (``budget_amounts``) of budgets (either int ID or :py:class:`~.Budget`
        instances) to Decimal amounts, ensure that the BudgetTransactions for
        this Transaction match those amounts.

        This method does NOT commit changes; it will modify database state
        and add the modifications to this object's session, but the calling
        code must commit changes.

        :param budget_amounts: Mapping of one or more Budgets to the amount of
          this Transaction allocated to that Budget. Keys may be either an int
          :py:attr:`~.Budget.id` or a :py:class:`~.Budget` instance, values must
          be a Decimal.
        :type budget_amounts: dict
        """
        assert isinstance(budget_amounts, type({}))
        assert len(budget_amounts) > 0
        logger.debug(
            'Setting budget amounts on %s to: %s', self, budget_amounts
        )
        sess = inspect(self).session
        # ensure keys are all Budget objects, not IDs
        for k in list(budget_amounts.keys()):
            if not isinstance(k, Budget):
                budg = sess.query(Budget).get(k)
                logger.debug('Swapping key from %s to %s', k, budg)
                budget_amounts[budg] = budget_amounts[k]
                del budget_amounts[k]
        # delete BudgetTransactions for any budget we're not using anymore,
        # or update amounts on any that are changed
        for btrans in self.budget_transactions:
            if btrans.budget not in budget_amounts.keys():
                logger.info('Removing %s from %s', btrans, self)
                sess.delete(btrans)
            elif btrans.amount != budget_amounts[btrans.budget]:
                logger.debug(
                    'Updating amount to %s on %s for %s',
                    budget_amounts[btrans.budget], btrans, self
                )
                btrans.amount = budget_amounts[btrans.budget]
        # finally, add any new ones that didn't exist before
        btrans = {bt.budget: bt for bt in self.budget_transactions}
        for budg in budget_amounts.keys():
            if budg not in btrans:
                bt = BudgetTransaction(
                    transaction=self,
                    budget=budg,
                    amount=budget_amounts[budg]
                )
                logger.debug('Adding %s to %s', bt, self)
                # SQLAlchemy 2.0 requires explicit add to session when
                # Transaction is already in a session; when called from
                # constructor, sess is None and cascade works implicitly
                if sess is not None:
                    sess.add(bt)
